﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.ComponentModel;
using System.Collections.ObjectModel;
using Buzz.MachineInterface;
using BuzzGUI.Interfaces;
using BuzzGUI.Common;
using System.Threading;
using System.Collections;

namespace WDE.EasyRec
{
    enum EasyRecState
    {
        Stop = 0,
        Record
    };

    [MachineDecl(Name = "WDE EasyRec", ShortName = "EasyRec", Author = "WDE", MaxTracks = 1)]
    public class EasyRec : IBuzzMachine
    {
        IBuzzMachineHost host;
        SampleData sampleData;
        int currentSlot = 0;

        public EasyRec(IBuzzMachineHost host)
        {
            this.host = host;
            sampleData = new SampleData();
        }

        int freezerState = (int)EasyRecState.Stop;

        [ParameterDecl(MaxValue = 1, DefValue = 0, ValueDescriptions = new String[] { "Stop", "Record" })]
        public int Mode
        {
            get { return freezerState; }
            set
            {
                switch ((int)value)
                {
                    case (int)EasyRecState.Record:
                        Application.Current.Dispatcher.BeginInvoke(new Action(() =>
                        {
                            sampleData.NewRecord();
                            host.MasterInfo.PosInTick = 0;
                        }));
                        break;
                    case (int)EasyRecState.Stop:
                        Application.Current.Dispatcher.BeginInvoke(new Action(() =>
                        {
                            if (!sampleData.IsEmpty())
                            {
                                saveToWavetable(currentSlot);
                                sampleData.Init();
                            }
                        }));
                        break;
                }

                freezerState = (int)value;
            }
        }

        public void Stop()
        {
            if (freezerState == (int)EasyRecState.Record)
            {
                foreach (IParameterGroup paramGroup in host.Machine.ParameterGroups)
                    foreach (IParameter param in paramGroup.Parameters)
                    {
                        if (param.Name == "Mode")
                        {
                            param.SetValue(0, (int)EasyRecState.Stop);
                        }
                    }
            }
        }

        public bool Work(Sample[] output, Sample[] input, int numsamples, WorkModes mode)
        {

            if (mode == WorkModes.WM_NOIO)
                return false;
            if (mode == WorkModes.WM_READ)                        // <thru>
                return true;

            bool ret = true; //Create sound

            switch (freezerState)
            {
                case (int)EasyRecState.Record:
                    {
                        if (Global.Buzz.Playing || Global.Buzz.Recording)
                        {
                            if (mode == WorkModes.WM_WRITE) // No audio or silence in input
                            {
                                sampleData.AppendSilence(numsamples);
                                for (int i = 0; i < numsamples; i++)
                                {
                                    output[i] = new Sample(0.0f, 0.0f);
                                }
                                return true;
                            }
                            else
                            {
                                sampleData.Append(input, numsamples);
                            }
                        }
                    }
                    break;

                case (int)EasyRecState.Stop:
                    {
                        if (mode == WorkModes.WM_WRITE)
                            return false;
                    }
                    break;
            }

            if (mode == WorkModes.WM_WRITE && !(Global.Buzz.Playing || Global.Buzz.Recording))
            {
                return false;
            }

            for (int i = 0; i < numsamples; i++)
            {
                output[i] = input[i];
            }
            return ret;
        }


        // actual machine ends here. the stuff below demonstrates some other features of the api.

        public IEnumerable<IMenuItem> Commands
        {
            get
            {

                yield return new MenuItemVM()
                {
                    Text = "Help",
                    Command = new SimpleCommand()
                    {
                        CanExecuteDelegate = p => true,
                        ExecuteDelegate = p => MessageBox.Show("Modes\n\nRecord: Start recording input audio when song is played.\nStop: Stop recording and copy audio to wavetable.")
                    }
                };

                yield return new MenuItemVM()
                {
                    Text = "About...",
                    Command = new SimpleCommand()
                    {
                        CanExecuteDelegate = p => true,
                        ExecuteDelegate = p => MessageBox.Show("Version 1.0.1 (C) 2017 WDE")
                    }
                };
            }
        }

        public void saveToWavetable(int slot)
        {
            var wt = host.Machine.Graph.Buzz.Song.Wavetable;
            WaveFormat wf;
            wf = WaveFormat.Float32;

            // write to wavetable
            int rootnote = BuzzNote.FromMIDINote(48);
            wt.AllocateWave(slot,
                                "",
                                "EasyRec",
                                (int)(sampleData.GetBuffer().Length),
                                wf,
                                true,
                                rootnote,
                                false,
                                true);
            IWaveLayer layer = wt.Waves[slot].Layers.Last();
            layer.SampleRate = Global.Buzz.SelectedAudioDriverSampleRate;

            float[] tmpBuffer = new float[sampleData.GetBuffer().Length];
            for (int i = 0; i < sampleData.GetBuffer().Length; i++)
            {
                tmpBuffer[i] = sampleData.GetBuffer()[i] / 32768.0f;
            }

            layer.SetDataAsFloat(tmpBuffer, 0, 2, 0, 0, (int)sampleData.GetBuffer().Length / 2); // Left
            layer.SetDataAsFloat(tmpBuffer, 1, 2, 1, 0, (int)sampleData.GetBuffer().Length / 2); // Right

            layer.InvalidateData();
        }

        internal void slotUpdated(int decAgain)
        {
            currentSlot = decAgain;
        }
    }

    public class MachineGUIFactory : IMachineGUIFactory { public IMachineGUI CreateGUI(IMachineGUIHost host) { return new EasyRecGUI(); } }
    public class EasyRecGUI : UserControl, IMachineGUI
    {
        IMachine machine;
        EasyRec EasyRecMachine;
        TextBox tb;
        ComboBox comboWT;

        // view model for machine list box items
        public class MachineVM
        {
            public IMachine Machine { get; private set; }
            public MachineVM(IMachine m) { Machine = m; }
            public override string ToString() { return Machine.Name; }
        }

        public ObservableCollection<MachineVM> Machines { get; private set; }

        public IMachine Machine
        {
            get { return machine; }
            set
            {
                if (machine != null)
                {
                    BindingOperations.ClearBinding(tb, TextBox.TextProperty);
                }

                machine = value;

                if (machine != null)
                {
                    EasyRecMachine = (EasyRec)machine.ManagedMachine;

                    tb.SetBinding(TextBox.TextProperty, new Binding("MachineState.Text") { Source = EasyRecMachine, Mode = BindingMode.TwoWay, UpdateSourceTrigger = UpdateSourceTrigger.PropertyChanged });
                }
            }
        }


        public EasyRecGUI()
        {
            Machines = new ObservableCollection<MachineVM>();

            TextBlock textBlock = new TextBlock() { Text = "Select Wavetable Slot", Margin = new Thickness(0, 0, 0, 4) };

            tb = new TextBox() { Margin = new Thickness(0, 0, 0, 4), AllowDrop = true };

            var sp = new StackPanel();

            sp.Children.Add(textBlock);

            this.Content = sp;
            comboWT = new ComboBox() { Margin = new Thickness(0, 0, 0, 4) };
            for (int i = 1; i < 201; i++)
            {
                comboWT.Items.Add(new TextBlock() { Text = i.ToString("X") });
            }
            comboWT.SelectedIndex = 0;
            comboWT.SelectionChanged += new SelectionChangedEventHandler(comboWT_SelectionChanged);

            sp.Children.Add(comboWT);

        }

        void comboWT_SelectionChanged(object sender, SelectionChangedEventArgs e)
        {
            string hex = ((TextBlock)comboWT.SelectedItem).Text;
            int decAgain = int.Parse(hex, System.Globalization.NumberStyles.HexNumber) - 1;
            if (EasyRecMachine != null)
                this.Dispatcher.Invoke(System.Windows.Threading.DispatcherPriority.Normal,
                 new Action(
                   delegate()
                   {
                       EasyRecMachine.slotUpdated(decAgain);
                   }
               ));
        }

        void bt_Click(object sender, RoutedEventArgs e)
        {
            string hex = ((TextBlock)comboWT.SelectedItem).Text;
            string message = "Are you sure you want to copy recorded audio to Wavetable slot " + hex + "?";
            string caption = "Copy audio to wavetable?";
            MessageBoxButton buttons = MessageBoxButton.YesNo;

            MessageBoxResult result = MessageBox.Show(message, caption, buttons);
            if (result == MessageBoxResult.Yes)
            {
                int decAgain = int.Parse(hex, System.Globalization.NumberStyles.HexNumber) - 1;
                if (EasyRecMachine != null)
                    this.Dispatcher.Invoke(System.Windows.Threading.DispatcherPriority.Normal,
                     new Action(
                       delegate()
                       {
                           EasyRecMachine.saveToWavetable(decAgain);
                       }
                   ));

            }
        }
    }
}
